@@ -70,6 +70,7 @@
 #include "xalloc.h"
 #include "all-io.h"
 #include "fileutils.h"
+#include "ttyutils.h"
 
 #include "logindefs.h"
 
@@ -109,7 +110,7 @@ struct login_context {
 	char		vcsan[VCS_PATH_MAX];
 #endif
 
-	char		thishost[MAXHOSTNAMELEN + 1];	/* this machine */
+	char		*thishost;			/* this machine */
 	char		*thisdomain;			/* this machine domain */
 	char		*hostname;			/* remote machine */
 	char		hostaddress[16];		/* remote address */
@@ -209,13 +210,13 @@ static void __attribute__ ((__noreturn__)) sleepexit(int eval)
 
 static const char *get_thishost(struct login_context *cxt, const char **domain)
 {
-	if (!*cxt->thishost) {
-		if (gethostname(cxt->thishost, sizeof(cxt->thishost))) {
+	if (!cxt->thishost) {
+		cxt->thishost = xgethostname();
+		if (!cxt->thishost) {
 			if (domain)
 				*domain = NULL;
 			return NULL;
 		}
-		cxt->thishost[sizeof(cxt->thishost) -1] = '\0';
 		cxt->thisdomain = strchr(cxt->thishost, '.');
 		if (cxt->thisdomain)
 			*cxt->thisdomain++ = '\0';
@@ -350,13 +351,12 @@ static void chown_tty(struct login_context *cxt)
  */
 static void init_tty(struct login_context *cxt)
 {
-	const char *p;
 	struct stat st;
 	struct termios tt, ttt;
 
 	cxt->tty_mode = (mode_t) getlogindefs_num("TTYPERM", TTY_MODE);
 
-	cxt->tty_path = ttyname(0);		/* libc calls istty() here */
+	get_terminal_name(&cxt->tty_path, &cxt->tty_name, &cxt->tty_number);
 
 	/*
 	 * In case login is suid it was possible to use a hardlink as stdin
@@ -375,18 +375,6 @@ static void init_tty(struct login_context *cxt)
 		sleepexit(EXIT_FAILURE);
 	}
 
-	if (strncmp(cxt->tty_path, "/dev/", 5) == 0)
-		cxt->tty_name = cxt->tty_path + 5;
-	else
-		cxt->tty_name = cxt->tty_path;
-
-	for (p = cxt->tty_name; p && *p; p++) {
-		if (isdigit(*p)) {
-			cxt->tty_number = p;
-			break;
-		}
-	}
-
 #ifdef LOGIN_CHOWN_VCS
 	if (cxt->tty_number) {
 		/* find names of Virtual Console devices, for later mode change */
@@ -518,7 +506,8 @@ static void log_lastlog(struct login_context *cxt)
 	if (fd < 0)
 		return;
 
-	lseek(fd, (off_t) cxt->pwd->pw_uid * sizeof(ll), SEEK_SET);
+	if (lseek(fd, (off_t) cxt->pwd->pw_uid * sizeof(ll), SEEK_SET) == -1)
+		goto done;
 
 	/*
 	 * Print last log message
@@ -536,7 +525,8 @@ static void log_lastlog(struct login_context *cxt)
 				printf(_("on %.*s\n"),
 				       (int)sizeof(ll.ll_line), ll.ll_line);
 		}
-		lseek(fd, (off_t) cxt->pwd->pw_uid * sizeof(ll), SEEK_SET);
+		if (lseek(fd, (off_t) cxt->pwd->pw_uid * sizeof(ll), SEEK_SET) == -1)
+			goto done;
 	}
 
 	memset((char *)&ll, 0, sizeof(ll));
@@ -551,7 +541,7 @@ static void log_lastlog(struct login_context *cxt)
 
 	if (write_all(fd, (char *)&ll, sizeof(ll)))
 		warn(_("write lastlog failed"));
-
+done:
 	close(fd);
 }
 
@@ -1031,13 +1021,17 @@ static int get_hushlogin_status(struct passwd *pwd)
 			gid_t egid = getegid();
 
 			sprintf(buf, "%s/%s", pwd->pw_dir, file);
-			setregid(-1, pwd->pw_gid);
-			setreuid(0, pwd->pw_uid);
-			ok = effective_access(buf, O_RDONLY) == 0;
-			setuid(0);	/* setreuid doesn't do it alone! */
-			setreuid(ruid, 0);
-			setregid(-1, egid);
 
+			if (setregid(-1, pwd->pw_gid) == 0 &&
+			    setreuid(0, pwd->pw_uid) == 0)
+				ok = effective_access(buf, O_RDONLY) == 0;
+
+			if (setuid(0) != 0 ||
+			    setreuid(ruid, 0) != 0 ||
+			    setregid(-1, egid) != 0) {
+				syslog(LOG_ALERT, _("hush login status: restore original IDs failed"));
+				exit(EXIT_FAILURE);
+			}
 			if (ok)
 				return 1;	/* enabled by user */
 		}
